"""Classes and functions for NVRTC (NVIDIA Runtime Compilation) using
ctypes."""

import ctypes
from typing import List, Tuple
from enum import Enum

from importlib_resources import files as importlib_resources_files


NVRTC_LIB = "libnvrtc.so.12"


def _find_libnvrtc() -> str:
    """Find the NVRTC shared lib in the nvidia.cuda_nvrtc package."""
    return str(importlib_resources_files("nvidia.cuda_nvrtc.lib").joinpath(NVRTC_LIB))


# Define the types.
class _nvrtc_Program(ctypes.Structure):
    """Opaque NVRTC program type."""

    _fields_ = []


nvrtc_Program = ctypes.POINTER(_nvrtc_Program)


def _define_nvrtc_types(nvrtc):
    """Define the function signatures."""
    nvrtc.nvrtcVersion.restype = ctypes.c_int
    nvrtc.nvrtcVersion.argtypes = [
        ctypes.POINTER(ctypes.c_int),
        ctypes.POINTER(ctypes.c_int),
    ]

    nvrtc.nvrtcGetErrorString.restype = ctypes.c_char_p
    nvrtc.nvrtcGetErrorString.argtypes = [ctypes.c_int]

    nvrtc.nvrtcCreateProgram.restype = ctypes.c_int
    nvrtc.nvrtcCreateProgram.argtypes = [
        ctypes.POINTER(nvrtc_Program),
        ctypes.c_char_p,
        ctypes.c_char_p,
        ctypes.c_int,
        ctypes.POINTER(ctypes.c_char_p),
        ctypes.POINTER(ctypes.c_char_p),
    ]

    nvrtc.nvrtcDestroyProgram.restype = ctypes.c_int
    nvrtc.nvrtcDestroyProgram.argtypes = [ctypes.POINTER(nvrtc_Program)]

    nvrtc.nvrtcCompileProgram.restype = ctypes.c_int
    nvrtc.nvrtcCompileProgram.argtypes = [
        nvrtc_Program,
        ctypes.c_int,
        ctypes.POINTER(ctypes.c_char_p),
    ]

    nvrtc.nvrtcGetPTXSize.restype = ctypes.c_int
    nvrtc.nvrtcGetPTXSize.argtypes = [nvrtc_Program, ctypes.POINTER(ctypes.c_size_t)]

    nvrtc.nvrtcGetPTX.restype = ctypes.c_int
    nvrtc.nvrtcGetPTX.argtypes = [nvrtc_Program, ctypes.c_char_p]

    nvrtc.nvrtcGetCUBINSize.restype = ctypes.c_int
    nvrtc.nvrtcGetCUBINSize.argtypes = [nvrtc_Program, ctypes.POINTER(ctypes.c_size_t)]

    nvrtc.nvrtcGetCUBIN.restype = ctypes.c_int
    nvrtc.nvrtcGetCUBIN.argtypes = [nvrtc_Program, ctypes.c_char_p]

    nvrtc.nvrtcGetProgramLogSize.restype = ctypes.c_int
    nvrtc.nvrtcGetProgramLogSize.argtypes = [
        nvrtc_Program,
        ctypes.POINTER(ctypes.c_size_t),
    ]

    nvrtc.nvrtcGetProgramLog.restype = ctypes.c_int
    nvrtc.nvrtcGetProgramLog.argtypes = [nvrtc_Program, ctypes.c_char_p]


try:
    lib_path = _find_libnvrtc()
except FileNotFoundError as e:
    raise ValueError(
        f"Could not find {NVRTC_LIB}. Did you install PyTorch with CUDA support?"
    ) from e
_nvrtc = ctypes.CDLL(lib_path)
_define_nvrtc_types(_nvrtc)


# Define the NVRTC error codes.
class NVRTCErrorCode(Enum):
    """NVRTC error codes."""

    NVRTC_SUCCESS = 0
    NVRTC_ERROR_OUT_OF_MEMORY = 1
    NVRTC_ERROR_PROGRAM_CREATION_FAILURE = 2
    NVRTC_ERROR_INVALID_INPUT = 3
    NVRTC_ERROR_INVALID_PROGRAM = 4
    NVRTC_ERROR_INVALID_OPTION = 5
    NVRTC_ERROR_COMPILATION = 6
    NVRTC_ERROR_BUILTIN_OPERATION_FAILURE = 7
    NVRTC_ERROR_NO_NAME_EXPRESSIONS_AFTER_COMPILATION = 8
    NVRTC_ERROR_NO_LOWERED_NAMES_BEFORE_COMPILATION = 9
    NVRTC_ERROR_NAME_EXPRESSION_NOT_VALID = 10
    NVRTC_ERROR_INTERNAL_ERROR = 11
    NVRTC_ERROR_TIME_FILE_WRITE_FAILED = 12


def version() -> Tuple[int, int]:
    """Return the NVRTC version as a (major, minor) tuple."""
    major = ctypes.c_int()
    minor = ctypes.c_int()
    result = _nvrtc.nvrtcVersion(ctypes.byref(major), ctypes.byref(minor))
    if result != 0:
        raise ValueError("Failed to get NVRTC version")
    return major.value, minor.value


def error_string(code: int) -> str:
    """Return the error string for the given NVRTC error code."""
    return _nvrtc.nvrtcGetErrorString(code).decode("utf-8")


class NVRTCError(Exception):
    """Exception raised for NVRTC errors."""

    def __init__(self, err: int):
        self.err = err
        super().__init__(error_string(err))


def _check(err: int):
    """Check error and raise an exception if not NVRTC_SUCCESS."""
    if err != NVRTCErrorCode.NVRTC_SUCCESS.value:
        raise NVRTCError(err)


class Program:
    """NVRTC program class for compiling CUDA code."""

    def __init__(
        self,
        src: str,
        name: str,
        headers: List[str] = None,
        include_names: List[str] = None,
    ):
        """Create a new NVRTC program."""
        if headers is None:
            headers = []
        if include_names is None:
            include_names = []
        self.program = nvrtc_Program()
        num_headers = len(headers)
        headers_arr = (ctypes.c_char_p * num_headers)()
        headers_arr[:] = [header.encode("utf-8") for header in headers]
        include_names_arr = (ctypes.c_char_p * len(include_names))()
        include_names_arr[:] = [name.encode("utf-8") for name in include_names]
        _check(
            _nvrtc.nvrtcCreateProgram(
                ctypes.byref(self.program),
                src.encode("utf-8"),
                name.encode("utf-8"),
                num_headers,
                headers_arr,
                include_names_arr,
            )
        )

    def __del__(self):
        _nvrtc.nvrtcDestroyProgram(ctypes.byref(self.program))

    def compile(self, options: List[str] = None):
        """Compile the program with the given options."""
        if options is None:
            options = []
        utf8_options = [option.encode("utf-8") for option in options]
        options = (ctypes.c_char_p * len(options))(*utf8_options)
        _check(_nvrtc.nvrtcCompileProgram(self.program, len(options), options))

    def ptx(self) -> str:
        """Return the PTX code as a string."""
        size = ctypes.c_size_t()
        _check(_nvrtc.nvrtcGetPTXSize(self.program, ctypes.byref(size)))
        ptx = ctypes.create_string_buffer(size.value)
        _check(_nvrtc.nvrtcGetPTX(self.program, ptx))
        return ptx.value.decode("utf-8")

    def cubin(self) -> bytes:
        """Return the CUBIN code as bytes."""
        size = ctypes.c_size_t()
        _check(_nvrtc.nvrtcGetCUBINSize(self.program, ctypes.byref(size)))
        cubin = ctypes.create_string_buffer(size.value)
        _check(_nvrtc.nvrtcGetCUBIN(self.program, cubin))
        return cubin.raw

    def log(self) -> str:
        """Return the compilation log as a string."""
        size = ctypes.c_size_t()
        _check(_nvrtc.nvrtcGetProgramLogSize(self.program, ctypes.byref(size)))
        log = ctypes.create_string_buffer(size.value)
        _check(_nvrtc.nvrtcGetProgramLog(self.program, log))
        return log.value.decode("utf-8")



"""Code generator for a custom strip loader class in CUDA / C++."""

from dataclasses import dataclass

from .gen_specs import GenSpecs
from .tensor import Tensor


@dataclass
class AsyncStripLoader(GenSpecs):
    """CUDA Code generator for custom strip loader classes.

    This class is used to generate custom strip loader classes that load
    data asynchronously from a global memory tensor to a shared memory
    tensor.
    """

    class_name: str
    smem_tensor: Tensor
    gmem_tensor: Tensor
    minor_axis: str
    major_axis_size: int
    minor_axis_size: int
    num_warps: int

    def __post_init__(self):
        """Normalize the minor axis name to upper-case."""
        self.minor_axis = self.minor_axis.upper()

    def generate(self) -> str:
        """Generate the C++ source code for the custom strip loader class."""
        smem_stride = self.smem_tensor.strides[self.minor_axis]
        gmem_stride = self.gmem_tensor.strides[self.minor_axis]
        params = self._gen_strip_loader_params()
        base_params = _make_args_list(smem_stride, gmem_stride, f"{params}::num_loads")
        base = f"spio::AsyncStripLoader<{base_params}>"
        return f"""
class {self.class_name} : public {base}
{{
    static constexpr int active_warps = {params}::active_warps;

    using Base = {base};
    using Base::Base;
}};
"""

    def _gen_strip_loader_params(self) -> str:
        pars = _make_args_list(
            self.major_axis_size, self.minor_axis_size, self.num_warps
        )
        return f"spio::StripLoaderParams<{pars}>"


def _make_args_list(*args):
    sep = ", "
    return sep.join(str(arg) for arg in args)



"""Code generator for custom folded-dimension template classes."""

from dataclasses import dataclass
from typing import Tuple

from .gen_specs import GenSpecs
from .dim import (
    dim_name_to_dim_or_fold_class_name,
    _format_fold_template_instance,
    _format_dim_class_name,
    _get_dim_name_and_stride,
)


@dataclass(frozen=True)
class Fold(GenSpecs):
    """CUDA Code generator for custom folded-dimension classes.

    This class defines a folding of a tensor dimension. The
    tensor dimension must already have been generated using DimSpec.
    """

    fold_name: str
    dim_name: str
    stride: int

    def __post_init__(self):
        """Normalize the fold and dimension names to upper-case."""
        object.__setattr__(self, "fold_name", self.fold_name.upper())
        object.__setattr__(self, "dim_name", self.dim_name.upper())

    def generate(self):
        dim_class_name = dim_name_to_dim_or_fold_class_name(self.dim_name)
        fold_template_instance = _format_fold_template_instance(
            dim_class_name, self.stride
        )
        fold_class_name = _format_dim_class_name(self.fold_name)
        return f"using {fold_class_name} = {fold_template_instance};\n"

    @property
    def dim_names(self) -> Tuple[str]:
        """Return the base dimension name, not the folded form.

        This ensures we don't create redundant dimension classes for already folded dimensions.
        """
        base_name, _ = _get_dim_name_and_stride(self.dim_name)
        return (base_name,)


def dim_header() -> str:
    """Return a C++ statement that includes the spio dim header.

    The header implements the C++ spio::Fold template class.
    """
    return '#include "spio/fold.h"'



"""Functions to transform and scan PyTorch modules with Spio modules."""

from typing import List

import torch
from torch.fx import symbolic_trace

from ..layers import Conv2dGw8
from ..util import get_full_name, logger_enabled
from ..kernels import Params


spio_modules_classes = [Conv2dGw8]


def transform(model: torch.nn.Module):
    """Transforms a PyTorch model to use Spio modules.

    Replaces any matching modules with their Spio counterparts.
    """
    return _transform(model)[0]


def _transform(model: torch.nn.Module):
    """Transform the model.

    Returns the transformed model and the number of modules replaced.
    """
    traced = symbolic_trace(model)
    num_replacements = 0
    modules_dict = dict(model.named_modules())
    nodes_to_replace = []

    # Collect nodes to be replaced
    for n in traced.graph.nodes:
        if n.op == "call_module" and n.target in modules_dict:
            module = modules_dict[n.target]
            for spio_module_class in spio_modules_classes:
                if spio_module_class.match(module):
                    nodes_to_replace.append(
                        (n, spio_module_class.from_torch_module(module))
                    )
                    break

    # Replace collected nodes
    for n, spio_module in nodes_to_replace:
        _replace_module(traced, n, spio_module)
        if logger_enabled:
            before = get_full_name(modules_dict[n.target].__class__)
            after = spio_module.__class__.__module__ + "." + str(spio_module)
            print(f"spio: replaced a {before} module with {after}")
        num_replacements += 1

    traced.delete_all_unused_submodules()
    traced.recompile()

    return traced, num_replacements


def scan_modules(model, *args) -> List[Params]:
    """Returns a list of Params for matched Spio modules.

    Scans a PyTorch model and returns a list of Params for the matched
    Spio modules.
    """
    traced = symbolic_trace(model)
    interpreter = _ScanInterpreter(traced)
    interpreter.run(*args)
    return interpreter.params_lst


def _replace_module(traced, n, spio_module):
    # Directly set the new Spio module in the model's named_modules dictionary
    if "." in n.target:
        parent_name, module_name = n.target.rsplit(".", 1)
        parent_module = dict(traced.named_modules())[parent_name]
    else:
        parent_module = traced
        module_name = n.target
    setattr(parent_module, module_name, spio_module)

    # Replace the node in the graph with the new Spio module
    with traced.graph.inserting_after(n):
        spio_node = traced.graph.call_module(n.target, args=n.args, kwargs=n.kwargs)
        n.replace_all_uses_with(spio_node)
    traced.graph.erase_node(n)


class _ScanInterpreter(torch.fx.Interpreter):
    """Interpreter for scanning modules in a PyTorch model."""

    def __init__(self, model):
        self.modules_dict = dict(model.named_modules())
        super().__init__(model)
        self._params_lst = []

    def call_module(self, target, args, kwargs):
        if target in self.modules_dict:
            module = self.modules_dict[target]
            for spio_module_class in spio_modules_classes:
                if spio_module_class.match(module):
                    params = spio_module_class.Params.from_torch_module(module, *args)
                    self.params_lst.append(params)
                    break
        return super().call_module(target, args, kwargs)

    @property
    def params_lst(self) -> List[Params]:
        """Return the list of parameters recorded so far."""
        return self._params_lst



"""Helper function for implementing kernel and operator unit tests."""

from typing import List, Type, Callable
import contextlib
import os

import torch

from ..compiler import compile_kernel_configs
from ..util.close import assert_all_close_with_acc_depth
from ..util.device_info import get_device_ordinal
from ..reflection import (
    get_kernel_reflection,
    get_function_reflection,
    get_layer_reflection,
)
from ..cuda.driver import get_device_attributes
from ..transform._transform import _transform as spio_transform
from ..kernels import Params, KernelFactory


# Pre-fill output tensors with this value to ensure every element is written by the kernel.
DEAD_VALUE = -42.0


def run_kernel_test(
    kernel_factory: KernelFactory,
    params: Params,
    device: str = "cuda",
    configs=None,
    **kernel_kwargs,
):
    """Run a test for an forward pass kernel."""
    kernel_name = kernel_factory.get_kernel_name(**kernel_kwargs)
    reflection = get_kernel_reflection(kernel_name)

    reference_function = reflection.reference
    reference_reflection = get_function_reflection(reference_function)

    args = reference_reflection.make_args(params, device=device)
    kernel_args = reflection.arrange_args(args)

    reference_args = _all_to_float(reference_reflection.arrange_args(args))
    torch_kwargs = reference_reflection.get_function_kwargs(params)

    device_idx = get_device_ordinal(device)
    device_attr = get_device_attributes(device_idx)

    with _precision_guard():
        torch_output = reference_function(*reference_args, **torch_kwargs)
    kernels = compile_kernel_configs(
        kernel_factory,
        params,
        configs=configs,
        device_attr=device_attr,
        **reflection.kwargs,
    )
    stats = reflection.stats(params, output_names=reflection.output_names)
    acc_depth = stats.accumulation_depths[0]
    output_name = reflection.output_names[0]
    output = args[output_name]
    for kernel in kernels:
        kernel.load()
        output.fill_(DEAD_VALUE)
        reflection.init_zeros(args)
        kernel(*kernel_args)
        assert_all_close_with_acc_depth(
            output, torch_output, msg=str(kernel), acc_depth=acc_depth
        )


def run_grad_kernel_test(
    kernel_factory: KernelFactory,
    params: Params,
    device: str = "cuda",
    configs=None,
    **kernel_kwargs,
):
    """Run a test for a backward pass kernel."""
    device_idx = get_device_ordinal(device)
    device_attr = get_device_attributes(device_idx)
    kernel_name = kernel_factory.get_kernel_name(**kernel_kwargs)
    reflection = get_kernel_reflection(kernel_name)
    args = reflection.make_args(params, device=device, training=True)

    reference_function = reflection.reference
    reference_reflection = get_function_reflection(reference_function)
    reference_args = _all_to_float(reference_reflection.arrange_args(args))
    reference_kwargs = reference_reflection.get_function_kwargs(params)

    grad_outputs = reflection.get_grad_output_args(args)

    with _precision_guard():
        output = reference_function(*reference_args, **reference_kwargs)
        num_grad_inputs = len(reflection.grad_input_names)
        ref_grads = {
            grad_input.name: torch.autograd.grad(
                output,
                args[grad_input.grad_of],
                grad_outputs,
                retain_graph=(idx < num_grad_inputs - 1),
            )[0]
            for idx, grad_input in enumerate(reflection.grad_input_names)
        }

    grad_input_names = [grad_name.name for grad_name in reflection.grad_input_names]
    grad_input_stats = [
        reflection.stats(params, output_names=[output_name])
        for output_name in grad_input_names
    ]
    acc_depths = [stats.accumulation_depths[0] for stats in grad_input_stats]

    kernel_args = reflection.arrange_args(args)
    kernels = compile_kernel_configs(
        kernel_factory,
        params,
        configs=configs,
        device_attr=device_attr,
        **kernel_kwargs,
    )
    for kernel in kernels:
        kernel.load()
        reflection.init_zeros(args)
        kernel(*kernel_args)
        for grad_name, acc_depth in zip(reflection.grad_input_names, acc_depths):
            grad = args[grad_name.name]
            ref_grad = ref_grads[grad_name.name]
            assert_all_close_with_acc_depth(
                grad, ref_grad, msg=str(kernel), acc_depth=acc_depth
            )


def run_opcheck_test(function, params, device="cuda"):
    """Check if a PyTorch custom operator was registered correctly."""
    reflection = get_function_reflection(function)
    args = reflection.make_args(params, device=device)
    function_args = reflection.arrange_args(args)
    function_kwargs = reflection.get_function_kwargs(params)
    torch.library.opcheck(
        function, function_args, function_kwargs, raise_exception=True
    )


def run_function_test(function, params, device="cuda"):
    """Run a test for the forward pass of a function."""
    reflection = get_function_reflection(function)

    reference_function = reflection.reference
    reference_reflection = get_function_reflection(reference_function)

    args = reference_reflection.make_args(params, device=device)
    function_args = reflection.arrange_args(args)
    function_args = _all_to_float(function_args)
    function_kwargs = reflection.get_function_kwargs(params)

    reference_args = _all_to_float(reference_reflection.arrange_args(args))
    reference_kwargs = reference_reflection.get_function_kwargs(params)

    with _precision_guard():
        reference_output = reference_function(*reference_args, **reference_kwargs)

    stats = reflection.stats(params, output_names=reflection.output_names)
    acc_depth = stats.accumulation_depths[0]

    with torch.autocast(device_type=device, dtype=torch.float16):
        output = function(*function_args, **function_kwargs)

    assert_all_close_with_acc_depth(
        output, reference_output, msg=str(params), acc_depth=acc_depth
    )


def run_grad_function_test(function: Callable, params: Params, device: str = "cuda"):
    """Run a test for a backward pass of a function."""
    reflection = get_function_reflection(function)
    args = reflection.make_args(params, device=device, training=True)
    function_args = reflection.arrange_args(args)
    function_args = _all_to_float(function_args)
    function_kwargs = reflection.get_function_kwargs(params)
    with torch.autocast(device_type=device, dtype=torch.float16):
        output = function(*function_args, **function_kwargs)

    reference_function = reflection.reference
    reference_reflection = get_function_reflection(reference_function)
    reference_args = reference_reflection.arrange_args(args)
    reference_kwargs = reference_reflection.get_function_kwargs(params)

    float_args = _all_to_float(reference_args)
    with _precision_guard():
        reference_outputs = reference_function(*float_args, **reference_kwargs)
    grad_outputs = list(reflection.make_grad_outputs(params).values())
    float_grad_outputs = _all_to_float(grad_outputs)
    grad_input_names = [f"grad_{name}" for name in reflection.args]
    grad_input_stats = [
        reflection.stats(params, output_names=[output_name])
        for output_name in grad_input_names
    ]
    acc_depths = [stats.accumulation_depths[0] for stats in grad_input_stats]
    for idx, (arg, float_arg) in enumerate(zip(function_args, float_args)):
        if arg is not None:
            not_last = idx < len(args) - 1
            grad = torch.autograd.grad(
                output, arg, *grad_outputs, retain_graph=not_last
            )
            with _precision_guard():
                reference_grad = torch.autograd.grad(
                    reference_outputs, float_arg, *float_grad_outputs, retain_graph=not_last
                )
            assert_all_close_with_acc_depth(
                grad[0], reference_grad[0], msg=str(params), acc_depth=acc_depths[idx]
            )


def run_layer_test(
    layer_cls: Type[torch.nn.Module],
    params: Params,
    device: str = "cuda",
    torchcompile: bool = False,
):
    """Run a test for a Spio layer."""
    reflection = get_layer_reflection(layer_cls)
    args = reflection.make_args(params, device=device)
    layer_args = reflection.arrange_args(args)
    layer_args = _all_to_float(layer_args)

    reference_layer_cls = reflection.reference
    reference_reflection = get_layer_reflection(reference_layer_cls)
    reference_args = reference_reflection.arrange_args(args)
    reference_kwargs = reference_reflection.get_function_kwargs(params)
    reference_args = _all_to_float(reference_args)

    reference_layer = reference_layer_cls(**reference_kwargs).to(device=device)
    with torch.no_grad():
        reference_layer.weight.copy_(args["weight"].float().clone().detach())
        if params.has_bias:
            reference_layer.bias.copy_(args["bias"].float().clone().detach())
    reference_layer = reference_layer.to(
        memory_format=reference_reflection.memory_format
    )

    reference_model = torch.nn.Sequential(reference_layer)
    with _precision_guard():
        reference_output = reference_model(*reference_args)

    layer, num_spio_modules = spio_transform(reference_model)
    assert num_spio_modules == 1, f"Expected 1 Spio module, matched {num_spio_modules}"

    if torchcompile:
        layer = torch.compile(layer)

    with torch.autocast(device_type=device, dtype=torch.float16):
        output = layer(*layer_args)

    stats = reflection.stats(params, output_names=reflection.output_names)
    acc_depth = stats.accumulation_depths[0]

    assert_all_close_with_acc_depth(
        output, reference_output, msg=str(params), acc_depth=acc_depth
    )


def _all_to_float(args: List[torch.Tensor]) -> List[torch.Tensor]:
    return [t.float() if t is not None else None for t in args]

@contextlib.contextmanager
def _precision_guard():
    # Disable TF32 control if explicitly requested
    if os.environ.get("SPIO_DISABLE_TF32", "1") == "0":
        yield
        return

    # Prefer PyTorch 2.9+ API; do not mix with old flags
    try:
        old_states = {
            "backends.fp32_precision": torch.backends.fp32_precision,
            "cuda.matmul.fp32_precision": torch.backends.cuda.matmul.fp32_precision,
            "cudnn.fp32_precision": torch.backends.cudnn.fp32_precision,
            "cudnn.conv.fp32_precision": torch.backends.cudnn.conv.fp32_precision,
            "cudnn.rnn.fp32_precision": torch.backends.cudnn.rnn.fp32_precision,
            "cudnn.benchmark": torch.backends.cudnn.benchmark,
        }
        torch.backends.fp32_precision = "ieee"
        torch.backends.cuda.matmul.fp32_precision = "ieee"
        torch.backends.cudnn.fp32_precision = "ieee"
        torch.backends.cudnn.conv.fp32_precision = "ieee"
        torch.backends.cudnn.rnn.fp32_precision = "ieee"
        torch.backends.cudnn.benchmark = False
        try:
            yield
        finally:
            torch.backends.fp32_precision = old_states["backends.fp32_precision"]
            torch.backends.cuda.matmul.fp32_precision = old_states["cuda.matmul.fp32_precision"]
            torch.backends.cudnn.fp32_precision = old_states["cudnn.fp32_precision"]
            torch.backends.cudnn.conv.fp32_precision = old_states["cudnn.conv.fp32_precision"]
            torch.backends.cudnn.rnn.fp32_precision = old_states["cudnn.rnn.fp32_precision"]
            torch.backends.cudnn.benchmark = old_states["cudnn.benchmark"]
        return
    except AttributeError:
        # Older PyTorch: fall back to legacy flags
        pass

    # Legacy path (pre-2.9)
    old_mm = torch.backends.cuda.matmul.allow_tf32
    old_cudnn = torch.backends.cudnn.allow_tf32
    old_bench = torch.backends.cudnn.benchmark
    try:
        torch.backends.cuda.matmul.allow_tf32 = False
        torch.backends.cudnn.allow_tf32 = False
        torch.backends.cudnn.benchmark = False
        try:
            torch.set_float32_matmul_precision("high")
        except Exception:
            pass
        yield
    finally:
        torch.backends.cuda.matmul.allow_tf32 = old_mm
        torch.backends.cudnn.allow_tf32 = old_cudnn
        torch.backends.cudnn.benchmark = old_bench



"""Test for NVCC compiler path detection."""

import pytest

import spio.compiler


@pytest.mark.skip(reason="NVCC support not requried by default.")
def test_nvcc():
    """Test NVCC compiler path is found."""
    assert spio.compiler.nvcc_full_path() is not None
